# -*- coding: utf-8 -*-
"""Saving utilities for YAML, JSON etc."""
import json
from collections import OrderedDict
from io import open
from typing import Dict, Any

import six
from ruamel.yaml import YAML, RoundTripRepresenter, add_representer, SafeRepresenter


class UnPrettyRTRepresenter(RoundTripRepresenter):
    def represent_none(self, data):
        # type: (Any) -> Any
        return self.represent_scalar(u'tag:yaml.org,2002:null', u'null')


class UnPrettySafeRepresenter(SafeRepresenter):
    def represent_none(self, data):
        # type: (Any) -> Any
        return self.represent_scalar(u'tag:yaml.org,2002:null', u'null')


class PrettySafeRepresenter(SafeRepresenter):
    def represent_none(self, data):
        # type: (Any) -> Any
        if len(self.represented_objects) == 0 and not self.serializer.use_explicit_start:
            # this will be open ended (although it is not yet)
            return self.represent_scalar(u'tag:yaml.org,2002:null', u'null')
        return self.represent_scalar(u'tag:yaml.org,2002:null', "")


UnPrettyRTRepresenter.add_representer(None, UnPrettyRTRepresenter.represent_none)
UnPrettySafeRepresenter.add_representer(None, UnPrettySafeRepresenter.represent_none)
PrettySafeRepresenter.add_representer(None, PrettySafeRepresenter.represent_none)


representers = {False: {False: UnPrettyRTRepresenter, True: RoundTripRepresenter}, True: {False: UnPrettySafeRepresenter, True: PrettySafeRepresenter}}


def save_str_to_file(string, path, encoding='utf-8'):
    # type: (str, str, str) -> None
    """Save string to file

    Args:
        string (str): String to save
        path (str): Path to file
        encoding (str): Encoding of file. Defaults to utf-8.

    Returns:
        None
    """
    with open(path, 'w', encoding=encoding) as f:
        if six.PY2:
            string = unicode(string)  # noqa: F821
        f.write(string)


def save_yaml(dictionary, path, encoding='utf-8', pretty=False, sortkeys=False):
    # type: (Dict, str, str, bool, bool) -> None
    """Save dictionary to YAML file preserving order if it is an OrderedDict

    Args:
        dictionary (Dict): Python dictionary to save
        path (str): Path to YAML file
        encoding (str): Encoding of file. Defaults to utf-8.
        pretty (bool): Whether to pretty print. Defaults to False.
        sortkeys (bool): Whether to sort dictionary keys. Defaults to False.

    Returns:
        None
    """
    with open(path, 'w', encoding=encoding) as f:
        representer = representers[sortkeys][pretty]
        yaml = YAML(typ='rt')
        yaml.Representer = representer
        add_representer(OrderedDict, representer.represent_dict, representer=representer)
        if pretty:
            yaml.indent(offset=2)
        else:
            yaml.default_flow_style = None
        yaml.representer.add_representer(type(None), representer.represent_none)
        yaml.dump(dictionary, f)


def save_json(dictionary, path, encoding='utf-8', pretty=False, sortkeys=False):
    # type: (Dict, str, str, bool, bool) -> None
    """Save dictionary to JSON file preserving order if it is an OrderedDict

    Args:
        dictionary (Dict): Python dictionary to save
        path (str): Path to JSON file
        encoding (str): Encoding of file. Defaults to utf-8.
        pretty (bool): Whether to pretty print. Defaults to False.
        sortkeys (bool): Whether to sort dictionary keys. Defaults to False.

    Returns:
        None
    """
    with open(path, 'w', encoding=encoding) as f:
        if pretty:
            indent = 2
            separators = (',', ': ')
        else:
            indent = None
            separators = (', ', ': ')
        if six.PY3:
            json.dump(dictionary, f, indent=indent, sort_keys=sortkeys, separators=separators)
        else:
            string = json.dumps(dictionary, ensure_ascii=False, indent=indent,
                                sort_keys=sortkeys, separators=separators)
            f.write(unicode(string))  # noqa: F821



import logging
from os import mkdir
from os.path import join
from shutil import rmtree

from hdx.utilities.downloader import DownloadError
from hdx.utilities.loader import load_json, load_file_to_str, load_yaml
from hdx.utilities.saver import save_json, save_str_to_file, save_yaml

logger = logging.getLogger(__name__)


class Retrieve(object):
    """Retrieve class which takes in a Download object and can either download, download and save or use previously
    downloaded and saved data. It also allows the use of a static fallback when downloading fails.

    Args:
        downloader (Download): Download object
        fallback_dir (str): Directory containing static fallback data
        saved_dir (str): Directory to save or load downloaded data
        temp_dir (str): Temporary directory for when data is not needed after downloading
        save (bool): Whether to save downloaded data. Defaults to False.
        use_saved (bool): Whether to use saved data. Defaults to False.
    """

    def __init__(self, downloader, fallback_dir, saved_dir, temp_dir, save=False, use_saved=False):
        self.downloader = downloader
        self.fallback_dir = fallback_dir
        self.saved_dir = saved_dir
        self.temp_dir = temp_dir
        self.save = save
        self.use_saved = use_saved
        if save:
            if use_saved:
                raise ValueError('Either the save or use_saved flags can be set to True!')
            rmtree(saved_dir, ignore_errors=True)
            mkdir(saved_dir)

    @staticmethod
    def get_url_logstr(url):
        # type: (str) -> str
        """Url string that will be logged. It is limited to 100 characters if necessary.

        Args:
            url (str): URL to download

        Returns:
            str: Url string to use in logs

        """
        if len(url) > 100:
            return '%s...' % url[:100]
        return url

    def retrieve_file(self, url, filename, logstr=None, fallback=False, **kwargs):
        """Retrieve file

        Args:
            url (str): URL to download
            filename (str): Filename to use for saved file
            logstr (Optional[str]): Text to use in log string to describe download. Defaults to filename.
            fallback (bool): Whether to use static fallback if download fails. Defaults to False.
            **kwargs: Parameters to pass to download_file call

        Returns:
            str: Path to downloaded file

        """
        if not logstr:
            logstr = filename
        if self.save:
            folder = self.saved_dir
        else:
            folder = self.temp_dir
        output_path = join(folder, filename)
        saved_path = join(self.saved_dir, filename)
        if self.use_saved:
            logger.info('Using saved %s in %s' % (logstr, saved_path))
            return saved_path
        else:
            try:
                logger.info('Downloading %s from %s into %s' % (logstr, self.get_url_logstr(url), output_path))
                return self.downloader.download_file(url, path=output_path, **kwargs)
            except DownloadError:
                if not fallback:
                    raise
                fallback_path = join(self.fallback_dir, filename)
                logger.exception('%s download failed, using static data %s!' % (logstr, fallback_path))
                return fallback_path

    def retrieve_text(self, url, filename, logstr=None, fallback=False, **kwargs):
        """Retrieve text

        Args:
            url (str): URL to download
            filename (str): Filename to use for saved file
            logstr (Optional[str]): Text to use in log string to describe download. Defaults to filename.
            fallback (bool): Whether to use static fallback if download fails. Defaults to False.
            **kwargs: Parameters to pass to download call

        Returns:
            Union[Dict,List]: The text from the file

        """
        if not logstr:
            logstr = filename
        saved_path = join(self.saved_dir, filename)
        if self.use_saved:
            logger.info('Using saved %s in %s' % (logstr, saved_path))
            text = load_file_to_str(saved_path)
        else:
            try:
                logger.info('Downloading %s from %s' % (logstr, self.get_url_logstr(url)))
                response = self.downloader.download(url, **kwargs)
                text = response.text
                if self.save:
                    logger.info('Saving %s in %s' % (logstr, saved_path))
                    save_str_to_file(text, saved_path)
            except DownloadError:
                if not fallback:
                    raise
                fallback_path = join(self.fallback_dir, filename)
                logger.exception('%s download failed, using static data %s!' % (logstr, fallback_path))
                text = load_file_to_str(fallback_path)
        return text

    def retrieve_yaml(self, url, filename, logstr=None, fallback=False, **kwargs):
        """Retrieve YAML

        Args:
            url (str): URL to download
            filename (str): Filename to use for saved file
            logstr (Optional[str]): Text to use in log string to describe download. Defaults to filename.
            fallback (bool): Whether to use static fallback if download fails. Defaults to False.
            **kwargs: Parameters to pass to download call

        Returns:
            Union[Dict,List]: The data from the YAML file

        """
        if not logstr:
            logstr = filename
        saved_path = join(self.saved_dir, filename)
        if self.use_saved:
            logger.info('Using saved %s in %s' % (logstr, saved_path))
            ryaml = load_yaml(saved_path)
        else:
            try:
                logger.info('Downloading %s from %s' % (logstr, self.get_url_logstr(url)))
                self.downloader.download(url, **kwargs)
                ryaml = self.downloader.get_yaml()
                if self.save:
                    logger.info('Saving %s in %s' % (logstr, saved_path))
                    save_yaml(ryaml, saved_path)
            except DownloadError:
                if not fallback:
                    raise
                fallback_path = join(self.fallback_dir, filename)
                logger.exception('%s download failed, using static data %s!' % (logstr, fallback_path))
                ryaml = load_yaml(fallback_path)
        return ryaml

    def retrieve_json(self, url, filename, logstr=None, fallback=False, **kwargs):
        """Retrieve JSON

        Args:
            url (str): URL to download
            filename (str): Filename to use for saved file
            logstr (Optional[str]): Text to use in log string to describe download. Defaults to filename.
            fallback (bool): Whether to use static fallback if download fails. Defaults to False.
            **kwargs: Parameters to pass to download call

        Returns:
            Union[Dict,List]: The data from the JSON file

        """
        if not logstr:
            logstr = filename
        saved_path = join(self.saved_dir, filename)
        if self.use_saved:
            logger.info('Using saved %s in %s' % (logstr, saved_path))
            rjson = load_json(saved_path)
        else:
            try:
                logger.info('Downloading %s from %s' % (logstr, self.get_url_logstr(url)))
                self.downloader.download(url, **kwargs)
                rjson = self.downloader.get_json()
                if self.save:
                    logger.info('Saving %s in %s' % (logstr, saved_path))
                    save_json(rjson, saved_path)
            except DownloadError:
                if not fallback:
                    raise
                fallback_path = join(self.fallback_dir, filename)
                logger.exception('%s download failed, using static data %s!' % (logstr, fallback_path))
                rjson = load_json(fallback_path)
        return rjson



# -*- coding: UTF-8 -*-
"""Retriever Tests"""
import copy
import random
import re
import string
from collections import OrderedDict
from os import remove, mkdir
from os.path import join, abspath
from shutil import rmtree

import pytest

from hdx.utilities.downloader import Download, DownloadError
from hdx.utilities.retriever import Retrieve
from hdx.utilities.useragent import UserAgent


class TestRetriever:
    retrieverfoldername = 'retriever'

    @pytest.fixture(scope='class', autouse=True)
    def useragent(self):
        UserAgent.set_global('test')
        yield
        UserAgent.clear_global()

    @pytest.fixture(scope='class')
    def retrieverfolder(self, fixturesfolder):
        return join(fixturesfolder, self.retrieverfoldername)

    @pytest.fixture(scope='class')
    def fallback_dir(self, retrieverfolder):
        return join(retrieverfolder, 'fallbacks')

    def test_retrieve_file(self, tmpdir, retrieverfolder, fallback_dir):
        tmpdir = str(tmpdir)
        saved_dir = join(tmpdir, 'saved')
        temp_dir = join(tmpdir, 'temp')
        rmtree(temp_dir, ignore_errors=True)
        mkdir(temp_dir)
        with Download() as downloader:
            with pytest.raises(ValueError):
                Retrieve(downloader, fallback_dir, saved_dir, temp_dir, save=True, use_saved=True)

            retriever = Retrieve(downloader, fallback_dir, saved_dir, temp_dir, save=False, use_saved=False)
            filename = 'test.txt'
            url = join(retrieverfolder, filename)
            path = retriever.retrieve_file(url, filename, logstr='test file', fallback=True)
            assert path == join(temp_dir, filename)
            path = retriever.retrieve_file('NOTEXIST', filename, logstr='test file', fallback=True)
            assert path == join(fallback_dir, filename)
            with pytest.raises(DownloadError):
                retriever.retrieve_file('NOTEXIST', filename, fallback=False)
            with pytest.raises(DownloadError):
                long_url = ''.join(random.SystemRandom().choice(string.ascii_uppercase + string.digits) for _ in range(150))
                retriever.retrieve_file(long_url, filename, fallback=False)
            text = retriever.retrieve_text(url, filename, logstr='test file', fallback=False)
            assert text == 'hello'
            text = retriever.retrieve_text('NOTEXIST', filename, logstr='test file', fallback=True)
            assert text == 'goodbye'
            with pytest.raises(DownloadError):
                retriever.retrieve_text('NOTEXIST', filename, fallback=False)
            filename = 'test.yaml'
            url = join(retrieverfolder, filename)
            data = retriever.retrieve_yaml(url, filename, logstr='test file', fallback=False)
            assert data['param_1'] == 'ABC'
            data = retriever.retrieve_yaml('NOTEXIST', filename, logstr='test file', fallback=True)
            assert data['param_1'] == 'XYZ'
            with pytest.raises(DownloadError):
                retriever.retrieve_yaml('NOTEXIST', filename, fallback=False)
            filename = 'test.json'
            url = join(retrieverfolder, filename)
            data = retriever.retrieve_json(url, filename, logstr='test file', fallback=False)
            assert data['my_param'] == 'abc'
            data = retriever.retrieve_json('NOTEXIST', filename, logstr='test file', fallback=True)
            assert data['my_param'] == 'xyz'
            with pytest.raises(DownloadError):
                retriever.retrieve_json('NOTEXIST', filename, fallback=False)

            retriever = Retrieve(downloader, fallback_dir, saved_dir, temp_dir, save=True, use_saved=False)
            filename = 'test.txt'
            url = join(retrieverfolder, filename)
            path = retriever.retrieve_file(url, filename, logstr='test file', fallback=True)
            assert path == join(saved_dir, filename)
            path = retriever.retrieve_file('NOTEXIST', filename, logstr='test file', fallback=True)
            assert path == join(fallback_dir, filename)
            with pytest.raises(DownloadError):
                retriever.retrieve_file('NOTEXIST', filename, fallback=False)
            text = retriever.retrieve_text(url, filename, logstr='test file', fallback=False)
            assert text == 'hello'
            text = retriever.retrieve_text('NOTEXIST', filename, logstr='test file', fallback=True)
            assert text == 'goodbye'
            with pytest.raises(DownloadError):
                retriever.retrieve_text('NOTEXIST', filename, fallback=False)
            filename = 'test.yaml'
            url = join(retrieverfolder, filename)
            data = retriever.retrieve_yaml(url, filename, logstr='test file', fallback=False)
            assert data['param_1'] == 'ABC'
            data = retriever.retrieve_yaml('NOTEXIST', filename, logstr='test file', fallback=True)
            assert data['param_1'] == 'XYZ'
            with pytest.raises(DownloadError):
                retriever.retrieve_yaml('NOTEXIST', filename, fallback=False)
            filename = 'test.json'
            url = join(retrieverfolder, filename)
            data = retriever.retrieve_json(url, filename, logstr='test file', fallback=False)
            assert data['my_param'] == 'abc'
            data = retriever.retrieve_json('NOTEXIST', filename, logstr='test file', fallback=True)
            assert data['my_param'] == 'xyz'
            with pytest.raises(DownloadError):
                retriever.retrieve_json('NOTEXIST', filename, fallback=False)

            retriever = Retrieve(downloader, fallback_dir, saved_dir, temp_dir, save=False, use_saved=True)
            filename = 'test.txt'
            url = join(retrieverfolder, filename)
            path = retriever.retrieve_file(url, filename, logstr='test file', fallback=True)
            assert path == join(saved_dir, filename)
            path = retriever.retrieve_file('NOTEXIST', filename, logstr='test file', fallback=True)
            assert path == join(saved_dir, filename)
            path = retriever.retrieve_file('NOTEXIST', filename, fallback=False)
            assert path == join(saved_dir, filename)
            text = retriever.retrieve_text(url, filename, logstr='test file', fallback=False)
            assert text == 'hello'
            text = retriever.retrieve_text('NOTEXIST', filename, logstr='test file', fallback=True)
            assert text == 'hello'
            text = retriever.retrieve_text('NOTEXIST', filename, fallback=False)
            assert text == 'hello'
            filename = 'test.yaml'
            url = join(retrieverfolder, filename)
            data = retriever.retrieve_yaml(url, filename, logstr='test file', fallback=False)
            assert data['param_1'] == 'ABC'
            data = retriever.retrieve_yaml('NOTEXIST', filename, logstr='test file', fallback=True)
            assert data['param_1'] == 'ABC'
            data = retriever.retrieve_yaml('NOTEXIST', filename, fallback=False)
            assert data['param_1'] == 'ABC'
            filename = 'test.json'
            url = join(retrieverfolder, filename)
            data = retriever.retrieve_json(url, filename, logstr='test file', fallback=False)
            assert data['my_param'] == 'abc'
            data = retriever.retrieve_json('NOTEXIST', filename, logstr='test file', fallback=True)
            assert data['my_param'] == 'abc'
            data = retriever.retrieve_json('NOTEXIST', filename, fallback=False)
            assert data['my_param'] == 'abc'

            retriever = Retrieve(downloader, fallback_dir, saved_dir, temp_dir, save=False, use_saved=True)
